/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.export;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.annotation.Configuration;
import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;

@Configuration
@ConfigurationProperties("ozgcloud.xdomea")
@Builder
@Getter
@Setter
@NoArgsConstructor
@AllArgsConstructor
public class XdomeaProperties implements Validator {

	private static final String PROP_BEHOERDENSCHLUESSEL_URI = "behoerdenschluesselUri";
	private static final String PROP_BEHOERDENSCHLUESSEL_VERSION = "behoerdenschluesselVersion";

	private String behoerdenschluessel;
	private String behoerdenschluesselUri;
	private String behoerdenschluesselVersion;

	@Override
	public boolean supports(Class<?> clazz) {
		return XdomeaProperties.class.isAssignableFrom(clazz);
	}

	@Override
	public void validate(Object target, Errors errors) {
		var properties = (XdomeaProperties) target;
		validateBehoerdenschluesselProperties(errors, properties);
	}

	private void validateBehoerdenschluesselProperties(Errors errors, XdomeaProperties properties) {
		if (StringUtils.isNotBlank(properties.getBehoerdenschluessel())) {
			validateNotBlank(PROP_BEHOERDENSCHLUESSEL_URI, properties, errors);
			validateNotBlank(PROP_BEHOERDENSCHLUESSEL_VERSION, properties, errors);
		}
	}

	private void validateNotBlank(String propertyName, XdomeaProperties properties, Errors errors) {
		getPropertyValue(propertyName, properties);
		if (StringUtils.isBlank(getPropertyValue(propertyName, properties))) {
			errors.rejectValue(propertyName, String.format("ozgcloud.xdomea.%s.empty", propertyName), String.format("%s must be set", propertyName));
		}
	}

	private static String getPropertyValue(String fieldName, XdomeaProperties properties) {
		try {
			return BeanUtils.getSimpleProperty(properties, fieldName);
		} catch (Exception e) {
			throw new TechnicalException("Property does not exist", e);
		}
	}
}
