/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.file;

import org.apache.commons.lang3.StringUtils;

import de.ozgcloud.archive.common.IdentifikationObjektTypeBuilder;
import de.ozgcloud.archive.common.VersionTypeBuilder;
import de.ozgcloud.archive.export.XdomeaMedienart;
import de.xoev.xdomea.AllgemeineMetadatenType;
import de.xoev.xdomea.DokumentType;
import de.xoev.xdomea.IdentifikationObjektType;
import de.xoev.xdomea.MediumCodeType;
import de.xoev.xdomea.VersionType;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class DokumentTypeBuilder {

	static final String VERSION_NUMMER = "1";

	private Long ordinalNumber;
	private String formEngineName;
	private OzgFile ozgFile;

	public static DokumentTypeBuilder builder() {
		return new DokumentTypeBuilder();
	}

	public DokumentTypeBuilder withOrdinalNumber(Long ordinalNumber) {
		this.ordinalNumber = ordinalNumber;
		return this;
	}

	public DokumentTypeBuilder withFormEngineName(String formEngineName) {
		this.formEngineName = formEngineName;
		return this;
	}

	public DokumentTypeBuilder withOzgFile(OzgFile ozgFile) {
		this.ozgFile = ozgFile;
		return this;
	}

	public DokumentType build() {
		var dokumentType = new DokumentType();
		dokumentType.setIdentifikation(createIdentifikation());
		dokumentType.setAllgemeineMetadaten(createAllgemeineMetadaten());
		dokumentType.getVersion().add(createVersionType());
		return dokumentType;
	}

	VersionType createVersionType() {
		return VersionTypeBuilder.builder()
				.withOzgFile(ozgFile)
				.withErsteller(formEngineName)
				.withSonstigerName(StringUtils.EMPTY)
				.build();
	}

	IdentifikationObjektType createIdentifikation() {
		return IdentifikationObjektTypeBuilder.builder()
				.withObjectID(ozgFile.getId().toString())
				.withOrdinalNumber(ordinalNumber)
				.build();
	}

	AllgemeineMetadatenType createAllgemeineMetadaten() {
		var allgemeineMetadaten = new AllgemeineMetadatenType();
		allgemeineMetadaten.setBetreff(ozgFile.getName());
		allgemeineMetadaten.setKennzeichen(StringUtils.EMPTY);
		allgemeineMetadaten.setBemerkung(StringUtils.EMPTY);
		allgemeineMetadaten.setMedium(createMediumCode());
		return allgemeineMetadaten;
	}

	private MediumCodeType createMediumCode() {
		var mediumCode = new MediumCodeType();
		mediumCode.setCode(XdomeaMedienart.ELEKTRONISCHES_DOKUMENT.getCode());
		return mediumCode;
	}
}
