/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.common.binaryfile;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.util.List;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.apilib.file.grpc.GrpcOzgCloudFileService;
import de.ozgcloud.archive.common.callcontext.CallContextProvider;
import de.ozgcloud.archive.file.OzgFile;
import de.ozgcloud.archive.file.OzgFileMapper;
import de.ozgcloud.archive.file.OzgFileTestFactory;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcBinaryFilesRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;

class BinaryFileRemoteServiceTest {

	@InjectMocks
	@Spy
	private BinaryFileRemoteService service;
	@Mock
	private BinaryFileServiceBlockingStub grpcService;
	@Mock
	private GrpcOzgCloudFileService grpcOzgCloudFileService;
	@Mock
	private OzgFileMapper ozgFileMapper;
	@Mock
	private FileIdMapper fileIdMapper;
	@Mock
	private CallContextProvider contextProvider;

	@Nested
	class TestGetFiles {

		private final GrpcFindFilesResponse response = GrpcFindFilesResponseTestFactory.create();
		private final GrpcBinaryFilesRequest request = GrpcBinaryFilesRequestTestFactory.create();
		private final FileId fileId1 = FileId.from(UUID.randomUUID().toString());
		private final FileId fileId2 = FileId.from(UUID.randomUUID().toString());

		@Mock
		private BinaryFileServiceBlockingStub grpcServiceWithInterceptor;

		@BeforeEach
		void mock() {
			doReturn(grpcServiceWithInterceptor).when(service).getGrpcServiceWithInterceptor();
			doReturn(request).when(service).buildBinaryFilesRequest(List.of(fileId1, fileId2));
			when(grpcServiceWithInterceptor.findBinaryFilesMetaData(request)).thenReturn(response);
		}

		@Test
		void shouldGetGrpcServiceWithInterceptor() {
			callService();

			verify(service).getGrpcServiceWithInterceptor();
		}

		@Test
		void shouldBuildBinaryFilesRequest() {
			callService();

			verify(service).buildBinaryFilesRequest(List.of(fileId1, fileId2));
		}

		@Test
		void shouldCallGrpcService() {
			callService();

			verify(grpcServiceWithInterceptor).findBinaryFilesMetaData(request);
		}

		@Test
		void shouldMapGrpcFindFilesRespone() {
			callService();

			verify(service).mapGrpcFindFilesRespone(response);
		}

		@Test
		void shouldReturnMappedOzgFiles() {
			var mappedOzgFile1 = OzgFileTestFactory.create();
			var mappedOzgFile2 = OzgFileTestFactory.create();
			doReturn(Stream.of(mappedOzgFile1, mappedOzgFile2)).when(service).mapGrpcFindFilesRespone(response);

			var ozgFiles = callService();

			assertThat(ozgFiles).containsExactlyInAnyOrder(mappedOzgFile1, mappedOzgFile2);
		}

		private Stream<OzgFile> callService() {
			return service.getFiles(List.of(fileId1, fileId2));
		}
	}

	@Nested
	class TestMapGrpcFindFilesRespone {

		@Test
		void shouldMapAllToOzgFile() {
			callService().toList();

			GrpcFindFilesResponseTestFactory.OZG_FILES.forEach(ozgFile -> verify(ozgFileMapper).fromGrpc(ozgFile));
		}

		@Test
		void shouldReturnMappedOzgFiles() {
			var mappedOzgFile1 = OzgFileTestFactory.create();
			var mappedOzgFile2 = OzgFileTestFactory.create();
			when(ozgFileMapper.fromGrpc(GrpcFindFilesResponseTestFactory.OZG_FILE_1)).thenReturn(mappedOzgFile1);
			when(ozgFileMapper.fromGrpc(GrpcFindFilesResponseTestFactory.OZG_FILE_2)).thenReturn(mappedOzgFile2);

			var ozgFiles = callService();

			assertThat(ozgFiles).containsExactlyInAnyOrder(mappedOzgFile1, mappedOzgFile2);
		}

		private Stream<OzgFile> callService() {
			return service.mapGrpcFindFilesRespone(GrpcFindFilesResponseTestFactory.create());
		}
	}

	@Nested
	class TestBuildBinaryFilesRequest {
		private final FileId fileId1 = FileId.from(GrpcBinaryFilesRequestTestFactory.ID_1);
		private final FileId fileId2 = FileId.from(GrpcBinaryFilesRequestTestFactory.ID_2);

		@Test
		void shouldBuildRequest() {
			var request = callService();

			assertThat(request).usingRecursiveComparison().isEqualTo(GrpcBinaryFilesRequestTestFactory.create());
		}

		private GrpcBinaryFilesRequest callService() {
			return service.buildBinaryFilesRequest(List.of(fileId1, fileId2));
		}
	}

	@Nested
	class TestGetFile {

		private final OzgCloudFileId ozgCloudFileId = OzgCloudFileTestFactory.ID;
		private final OzgCloudFile ozgCloudFile = OzgCloudFileTestFactory.create();

		@BeforeEach
		void mockFileIdMapper() {
			when(fileIdMapper.toOzgCloudFileId(OzgFileTestFactory.ID)).thenReturn(ozgCloudFileId);
		}

		@Test
		void shouldCallFileIdMapper() {
			callService();

			verify(fileIdMapper).toOzgCloudFileId(OzgFileTestFactory.ID);
		}

		@Test
		void shouldCallGrpcOzgCloudFileService() {
			callService();

			verify(grpcOzgCloudFileService).getFile(ozgCloudFileId);
		}

		@Test
		void shouldMapOzgCloudFile() {
			when(grpcOzgCloudFileService.getFile(ozgCloudFileId)).thenReturn(ozgCloudFile);

			callService();

			verify(ozgFileMapper).fromOzgCloudFile(ozgCloudFile);
		}

		@Test
		void shouldReturnMappedOzgFile() {
			var mappedOzgFile = OzgFileTestFactory.create();
			when(grpcOzgCloudFileService.getFile(ozgCloudFileId)).thenReturn(ozgCloudFile);
			when(ozgFileMapper.fromOzgCloudFile(ozgCloudFile)).thenReturn(mappedOzgFile);

			var ozgFile = callService();

			assertThat(ozgFile).isEqualTo(mappedOzgFile);
		}

		private OzgFile callService() {
			return service.getFile(OzgFileTestFactory.ID);
		}
	}

	@Nested
	class TestWriteFileContent {

		private final FileId fileId = OzgFileTestFactory.ID;
		private final OzgCloudFileId ozgCloudFileId = OzgCloudFileTestFactory.ID;

		@Mock
		private OutputStream outputStream;

		@Test
		void shouldCallFileIdMapper() {
			callService();

			verify(fileIdMapper).toOzgCloudFileId(fileId);
		}

		@Test
		void shouldCallGrpcOzgCloudFileService() {
			when(fileIdMapper.toOzgCloudFileId(fileId)).thenReturn(ozgCloudFileId);

			callService();

			verify(grpcOzgCloudFileService).writeFileDataToStream(ozgCloudFileId, outputStream);
		}

		private void callService() {
			service.writeFileContent(fileId, outputStream);
		}
	}

	@Nested
	class TestGetGrpcServiceWithInterceptor {

		private MockedConstruction<OzgCloudCallContextAttachingInterceptor> mockedConstructionInterceptor;
		private OzgCloudCallContextProvider argumentContextProvider;
		@Mock
		private OzgCloudCallContextAttachingInterceptor interceptor;

		@BeforeEach
		void setUpMocks() {
			mockedConstructionInterceptor = mockConstruction(OzgCloudCallContextAttachingInterceptor.class, (mock, context) -> {
				argumentContextProvider = (OzgCloudCallContextProvider) context.arguments().get(0);
				interceptor = mock;
			});
		}

		@AfterEach
		void cleanUp() {
			mockedConstructionInterceptor.close();
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptor() {
			callService();

			assertThat(mockedConstructionInterceptor.constructed()).hasSize(1);
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptorWithContextProvider() {
			callService();

			assertThat(argumentContextProvider).isEqualTo(contextProvider);
		}

		@Test
		void shouldSetInterceptor() {
			callService();

			verify(grpcService).withInterceptors(interceptor);
		}

		@Test
		void shouldReturnServiceStub() {
			when(grpcService.withInterceptors(any(OzgCloudCallContextAttachingInterceptor.class))).thenReturn(grpcService);

			var returnedServiceStub = callService();

			assertThat(returnedServiceStub).isEqualTo(grpcService);
		}

		private BinaryFileServiceBlockingStub callService() {
			return service.getGrpcServiceWithInterceptor();
		}
	}
}
