package de.ozgcloud.archive.common.xta;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;

import javax.xml.parsers.DocumentBuilder;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.core.io.ClassPathResource;
import org.w3c.dom.Document;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.xta.client.model.XtaFile;
import lombok.SneakyThrows;

class XtaFileParserTest {

	@Spy
	@InjectMocks
	private XtaFileParser parser;

	@Mock
	private DocumentBuilder documentBuilder;

	@DisplayName("Is import confirmation")
	@Nested
	class TestIsType {

		private final XtaFile file = XtaFileTestFactory.create();

		@Test
		@SneakyThrows
		void shouldCallParse() {
			isType();

			verify(parser).parse(file);
		}

		@Test
		@SneakyThrows
		void shouldReturnTrueOnMatchingTag() {
			doReturn(AbgabeImportBestaetigen0402TestFactory.createAsDocument()).when(parser).parse(file);

			var isImportConfirmation = isType();

			assertThat(isImportConfirmation).isTrue();
		}

		@Test
		@SneakyThrows
		void shouldReturnFalseOnNonMatchingTag() {
			doReturn(DocumentTestFactory.create()).when(parser).parse(file);

			var isImportConfirmation = isType();

			assertThat(isImportConfirmation).isFalse();
		}

		@Test
		@SneakyThrows
		void shouldReturnFalseOnException() {
			doThrow(TechnicalException.class).when(parser).parse(any());

			var isType = isType();

			assertThat(isType).isFalse();
		}

		private boolean isType() {
			return parser.isType(file, XtaFileType.IMPORT_CONFIRMATION);
		}
	}

	@Nested
	class TestParse {

		private final XtaFile file = XtaFileTestFactory.create();
		private final Document document = AbgabeImportBestaetigen0402TestFactory.createAsDocument();

		@Captor
		private ArgumentCaptor<InputStream> inputStreamCaptor;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(documentBuilder.parse(any(InputStream.class))).thenReturn(document);
		}

		@Test
		@SneakyThrows
		void shouldCallDocumentBuilder() {
			var expectedBytes = new ClassPathResource(XtaFileTestFactory.DUMMY_XML_FILE_PATH).getContentAsByteArray();

			parse();

			verify(documentBuilder).parse(inputStreamCaptor.capture());
			assertThat(inputStreamCaptor.getValue().readAllBytes()).isEqualTo(expectedBytes);
		}

		@Test
		@SneakyThrows
		void shouldReturnDocument() {
			var result = parse();

			assertThat(result).isEqualTo(document);
		}

		@SneakyThrows
		private Document parse() {
			return parser.parse(file);
		}
	}
}