package de.ozgcloud.archive.common.xta;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.function.Consumer;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.archive.common.command.CommandService;
import de.ozgcloud.archive.vorgang.XdomeaXtaFileTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.xdomea.XdomeaXtaMessageCreator;
import lombok.SneakyThrows;

class XtaServiceTest {

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaClient xtaClient;

	@Nested
	class TestSendXdomeaFile {

		private final XtaFile xdomeaFile = XdomeaXtaFileTestFactory.create();
		@Mock
		private XtaMessage xtaMessage;

		@BeforeEach
		void mock() {
			doReturn(xtaMessage).when(service).createMessage(xdomeaFile);
			doReturn(xtaClient).when(service).getXtaClient();
		}

		@Test
		void shouldGetXtaClient() {
			sendXdomeaFile();

			verify(service).getXtaClient();
		}

		@Test
		@SneakyThrows
		void shouldCallCreateMessage() {
			sendXdomeaFile();

			verify(service).createMessage(xdomeaFile);
		}

		@Test
		@SneakyThrows
		void shouldSendMessage() {
			sendXdomeaFile();

			verify(xtaClient).sendMessage(xtaMessage);
		}

		@Test
		@SneakyThrows
		void shouldThrowTechnicalException() {
			when(xtaClient.sendMessage(any())).thenThrow(XtaClientException.class);

			assertThrows(TechnicalException.class, () -> sendXdomeaFile());
		}

		private void sendXdomeaFile() {
			service.sendXdomeaFile(xdomeaFile);
		}
	}

	@Nested
	class TestCreateMessage {
		private final XtaFile xdomeaFile = XdomeaXtaFileTestFactory.create();

		private MockedStatic<XdomeaXtaMessageCreator> mockedStaticXdomeaXtaMessageCreator;

		@Mock
		private XdomeaXtaMessageCreator xdomeaXtaMessageCreator;
		@Mock
		private XtaMessage xtaMessage;

		@BeforeEach
		void setUpMock() {
			mockedStaticXdomeaXtaMessageCreator = mockStatic(XdomeaXtaMessageCreator.class);
			mockedStaticXdomeaXtaMessageCreator.when(XdomeaXtaMessageCreator::createInstance).thenReturn(xdomeaXtaMessageCreator);
		}

		@AfterEach
		void cleanUp() {
			mockedStaticXdomeaXtaMessageCreator.close();
		}

		@Test
		void shouldCreateXdomeaXtaMessageCreator() {
			createMessage();

			mockedStaticXdomeaXtaMessageCreator.verify(XdomeaXtaMessageCreator::createInstance);
		}

		@Test
		@SneakyThrows
		void shouldCallCreateMessage() {
			createMessage();

			verify(xdomeaXtaMessageCreator).createMessage(xdomeaFile);
		}

		@Test
		@SneakyThrows
		void shouldReturnMessage() {
			when(xdomeaXtaMessageCreator.createMessage(any())).thenReturn(xtaMessage);

			var message = createMessage();

			assertThat(message).isEqualTo(xtaMessage);
		}

		@Test
		@SneakyThrows
		void shouldThrowTechnicalException() {
			when(xdomeaXtaMessageCreator.createMessage(any())).thenThrow(XtaClientException.class);

			assertThrows(TechnicalException.class, () -> createMessage());
		}

		private XtaMessage createMessage() {
			return service.createMessage(xdomeaFile);
		}
	}

	@Nested
	class TestGetXtaClient {

		private XtaService xtaService;

		@Mock
		private XtaFileHelper xtaFileHelper;

		@Mock
		private CommandService commandService;

		@Mock
		private ApplicationEventPublisher eventPublisher;

		@Nested
		class TestOnConfiguredXtaClient {

			@BeforeEach
			void givenXtaClientConfigured() {
				xtaService = spy(new XtaService(Optional.of(xtaClient)));
			}

			@Test
			void shouldReturnXtaClient() {
				var returnedXtaClient = getXtaClient();

				assertThat(returnedXtaClient).isSameAs(xtaClient);
			}

		}

		@Nested
		class TestOnUnconfiguredXtaClient {

			@BeforeEach
			void givenXtaClientNotConfigured() {
				xtaService = spy(new XtaService(Optional.empty()));
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> getXtaClient());
			}
		}

		private XtaClient getXtaClient() {
			return xtaService.getXtaClient();
		}
	}

	@DisplayName("Fetch messages")
	@Nested
	class TestFetchMessages {

		@Mock
		private Consumer<XtaMessage> messageConsumer;

		@BeforeEach
		void mock() {
			doReturn(xtaClient).when(service).getXtaClient();
		}

		@Test
		void shouldCallGetXtaClient() {
			fetchMessages();

			verify(service).getXtaClient();
		}

		@SneakyThrows
		@Test
		void shouldCallClient() {
			fetchMessages();

			verify(xtaClient).fetchMessages(messageConsumer);
		}

		@SneakyThrows
		@Test
		void shouldThrowException() {
			when(xtaClient.fetchMessages(any())).thenThrow(new XtaClientException("error-message"));

			assertThatThrownBy(this::fetchMessages).isInstanceOf(TechnicalException.class);
		}

		private void fetchMessages() {
			service.fetchMessages(messageConsumer);
		}
	}

}
